@@ -35,6 +35,8 @@ struct proxy_interface;
 #define IWD_STATION_INTERFACE            "net.connman.iwd.Station"
 #define IWD_STATION_DIAGNOSTIC_INTERFACE "net.connman.iwd.StationDiagnostic"
 #define IWD_AP_DIAGNOSTIC_INTERFACE      "net.connman.iwd.AccessPointDiagnostic"
+#define IWD_DAEMON_INTERFACE             "net.connman.iwd.Daemon"
+#define IWD_DPP_INTERFACE                "net.connman.iwd.DeviceProvisioning"
 
 typedef bool (*proxy_property_match_func_t) (const void *a, const void *b);
 
@@ -117,10 +119,14 @@ struct interface_type_desc {
 } __attribute__((aligned(8)));
 
 #define INTERFACE_TYPE(interface, init, exit)				\
+	_Pragma("GCC diagnostic push")					\
+	_Pragma("GCC diagnostic ignored \"-Wattributes\"")		\
 	static struct interface_type_desc __interface_type_ ## interface\
-		__attribute__((used, section("__interface"), aligned(8))) = {\
+		__attribute__((used, retain, section("__interface"),	\
+					aligned(8))) = {		\
 			#interface, init, exit				\
 		};							\
+	_Pragma("GCC diagnostic pop")
 
 bool dbus_proxy_init(void);
 bool dbus_proxy_exit(void);
